import asyncio
import contextlib
import logging
import multiprocessing as mp
from functools import partial
from logging import DEBUG, LogRecord, getLogger
from logging.handlers import QueueHandler
from multiprocessing.context import BaseContext
from queue import Queue
from typing import Optional, cast

__all__ = ['MultiprocessingLogging']


def example_func() -> None:
    '''Used in doctest, defined here to be picklable.'''
    logger = logging.getLogger(__name__)
    logger.warning('warning from another process')


@contextlib.asynccontextmanager
async def MultiprocessingLogging(mp_context: Optional[BaseContext] = None):
    '''Collect logging from other processes in the main process.

    Example:

    A function to be executed in another process is example_func(), which is defined
    outside of the docstring because it must be picklable.

    Define the main asynchronous function, in which MultiprocessingLogging is used.

    >>> async def main():
    ...     from concurrent.futures import ProcessPoolExecutor
    ...
    ...     # Start MultiprocessingLogging.
    ...     async with MultiprocessingLogging() as initializer:
    ...
    ...         # The initializer is given to ProcessPoolExecutor.
    ...         with ProcessPoolExecutor(initializer=initializer) as executor:
    ...
    ...             # In another process, execute example_func(), which logs a warning.
    ...             future = executor.submit(example_func)
    ...
    ...             # Wait until the function returns.
    ...             future.result()

    When the main() is executed, the warning in example_func() is received in the
    main process.

    To confirm that it works in this example, we add a queue handler to the
    current logger.

    >>> queue = Queue()
    >>> handler = QueueHandler(queue)
    >>> logger = logging.getLogger(__name__)
    >>> logger.addHandler(handler)

    Run the main function.

    >>> asyncio.run(main())

    Check the loggings in the queue.

    >>> log_record = queue.get()
    >>> log_record.getMessage()
    'warning from another process'

    Remove the queue handler added for the example.

    >>> logger.removeHandler(handler)


    Reference: "Logging to a single file from multiple processes" (Logging Cookbook)
    https://docs.python.org/3/howto/logging-cookbook.html#logging-to-a-single-file-from-multiple-processes

    '''

    mp_context = mp_context or mp.get_context()
    queue = cast(Queue[LogRecord | None], mp_context.Queue())
    initializer = partial(_initializer, queue)

    async def _listen() -> None:
        '''Receive loggings from other processes and handle them in the current process.'''
        while (record := await asyncio.to_thread(queue.get)) is not None:
            logger = getLogger(record.name)
            if logger.getEffectiveLevel() <= record.levelno:
                logger.handle(record)

    task = asyncio.create_task(_listen())

    try:
        yield initializer
    finally:
        await asyncio.to_thread(queue.put, None)
        await task


def _initializer(queue: Queue[LogRecord]) -> None:
    '''An initializer of ProcessPoolExecutor.'''
    handler = QueueHandler(queue)
    logger = getLogger()
    logger.setLevel(DEBUG)
    logger.addHandler(handler)



from nextline.plugin.spec import Context, hookimpl


class RunNoRegistrar:
    @hookimpl
    async def on_initialize_run(self, context: Context) -> None:
        assert context.run_arg
        await context.pubsub.publish('run_no', context.run_arg.run_no)



from logging import getLogger

from nextline import events
from nextline.plugin.spec import Context, hookimpl


class OnEvent:
    @hookimpl
    async def on_event_in_process(self, context: Context, event: events.Event) -> None:
        ahook = context.hook.ahook
        match event:
            case events.OnStartTrace():
                await ahook.on_start_trace(context=context, event=event)
            case events.OnEndTrace():
                await ahook.on_end_trace(context=context, event=event)
            case events.OnStartTraceCall():
                await ahook.on_start_trace_call(context=context, event=event)
            case events.OnEndTraceCall():
                await ahook.on_end_trace_call(context=context, event=event)
            case events.OnStartCmdloop():
                await ahook.on_start_cmdloop(context=context, event=event)
            case events.OnEndCmdloop():
                await ahook.on_end_cmdloop(context=context, event=event)
            case events.OnStartPrompt():
                await ahook.on_start_prompt(context=context, event=event)
            case events.OnEndPrompt():
                await ahook.on_end_prompt(context=context, event=event)
            case events.OnWriteStdout():
                await ahook.on_write_stdout(context=context, event=event)
            case _:
                logger = getLogger(__name__)
                logger.warning(f'Unknown event: {event!r}')



from io import TextIOWrapper
from logging import getLogger
from typing import Optional, Protocol


class PromptFuncType(Protocol):
    '''Prompt the user with text for a command.'''

    def __call__(self, text: str) -> str:
        ...


class StdInOut(TextIOWrapper):
    '''A stdin and stdout interface to prompt_func(text: str) -> str.

    Example:

    Define a prompt_func() that receives the prompt text and returns the user command.
    >>> def prompt_func(text: str) -> str:
    ...     return f'Hi, I am the user. You said: {text!r}'

    In practice, the prompt_func() would prompt the user for a command and be blocked
    until the user responds.  Here, we just return a string.

    Initialize with the prompt_func().
    >>> stdio = StdInOut(prompt_func=prompt_func)

    The `stdio` is to be given as both stdin and stdout to Pdb, or any other
    Python object that can be operated via stdout.write() and stdin.readline().

    In this example, we will call stdio.write() and stdio.readline() directly.

    Write the prompt text to stdout.
    >>> _ = stdio.write('Hello, I am Pdb. ')

    Wait for the user response.
    >>> stdio.readline()
    "Hi, I am the user. You said: 'Hello, I am Pdb. '"

    '''

    def __init__(
        self, prompt_func: PromptFuncType, prompt_end: Optional[str] = None
    ) -> None:
        # To be matched to the end of prompt text if not None.
        self.prompt_end = prompt_end

        self._prompt = prompt_func
        self._prompt_text = ''

    def write(self, s: str) -> int:
        '''Print prompt or other text.

        For example, Pdb calls this method as stdout.write(pdb.prompt).
        '''
        self._prompt_text += s
        return len(s)

    def flush(self) -> None:
        '''To replace stdout.flush()'''
        pass

    def readline(self) -> str:  # type: ignore
        '''Call the prompt_func() with the text given to write() and return the result.

        For example, Pdb calls this method as stdin.readline() for user command.
        '''

        prompt_text = self._prompt_text

        logger = getLogger(__name__)

        logger.debug(f'Prompt text: {prompt_text!r}')

        if not self.prompt_end:
            logger.debug(f'Prompt end is not set: {self.prompt_end!r}')
        else:
            try:
                assert prompt_text.endswith(self.prompt_end)
            except AssertionError:
                msg = f'{prompt_text!r} does not end with {self.prompt_end!r}'
                logger.exception(msg)
                raise

        self._prompt_text = ''

        command = self._prompt(text=prompt_text)

        return command



import threading
from asyncio import Task
from collections.abc import Generator, Iterable, Iterator
from functools import lru_cache, partial
from logging import getLogger
from threading import Thread
from typing import Optional

from apluggy import PluginManager, contextmanager

from nextline.spawned.plugin.spec import hookimpl
from nextline.spawned.types import TraceArgs
from nextline.utils import current_task_or_thread, match_any

from . import _script


class FilterByModuleName:
    '''Skip Python modules with names that match any of the patterns.'''

    @hookimpl
    def init(self, modules_to_skip: Iterable[str]) -> None:
        self._patterns = frozenset(modules_to_skip)

        # NOTE: Use lru_cache() as match_any() is slow
        self._match_any_ = lru_cache(partial(match_any, patterns=modules_to_skip))

    @hookimpl
    def filter(self, trace_args: TraceArgs) -> bool | None:
        frame = trace_args[0]
        module_name = frame.f_globals.get('__name__')
        matched = self._match_any_(module_name)
        return matched or None


class FilterLambda:
    '''Skip lambda functions.'''

    @hookimpl
    def filter(self, trace_args: TraceArgs) -> bool | None:
        frame = trace_args[0]
        func_name = frame.f_code.co_name
        return func_name == '<lambda>' or None


class FilterMainScript:
    '''Accept only the main script.'''

    @hookimpl
    def filter(self, trace_args: TraceArgs) -> bool | None:
        frame = trace_args[0]
        module_name = frame.f_globals.get('__name__')
        if _script.__name__ == module_name:
            return False
        return True


class FilerByModule:
    '''Accept the first module and modules ever in the cmdloop() context.'''

    def __init__(self) -> None:
        self._modules_to_trace = set[str]()
        self._first_module_added = False
        self._entering_thread: Optional[Thread] = None
        self._traced_tasks_and_threads = set[Task | Thread]()
        self._logger = getLogger(__name__)

    @hookimpl
    def init(self, hook: PluginManager) -> None:
        self._hook = hook

    @hookimpl
    @contextmanager
    def context(self) -> Iterator[None]:
        self._entering_thread = threading.current_thread()
        msg = f'{self.__class__.__name__}: entering thread {self._entering_thread}'
        self._logger.info(msg)
        yield

    @hookimpl(trylast=True)
    def filter(self, trace_args: TraceArgs) -> bool | None:
        # Accept the first module in the main thread.
        if not self._first_module_added:
            if self._entering_thread == threading.current_thread():
                self._add(trace_args)
                self._first_module_added = True

        # Accept tasks and threads once they are accepted.
        task_or_thread = current_task_or_thread()
        if task_or_thread in self._traced_tasks_and_threads:
            return None

        # Accept if the module was in the cmdloop() in other tasks or threads.
        if self._to_trace(trace_args):
            self._traced_tasks_and_threads.add(task_or_thread)
            return None

        # Reject else.
        return True

    @hookimpl
    @contextmanager
    def on_cmdloop(self) -> Generator[None, str, None]:
        trace_args = self._hook.hook.current_trace_args()
        self._add(trace_args)
        yield

    def _add(self, trace_args: TraceArgs) -> None:
        frame, _, _ = trace_args
        module_name = frame.f_globals.get('__name__')
        if module_name is None:
            return
        if module_name in self._modules_to_trace:
            return
        self._modules_to_trace.add(module_name)
        msg = f'{self.__class__.__name__}: added {module_name!r}'
        self._logger.info(msg)

    def _to_trace(self, trace_args: TraceArgs) -> bool:
        frame, _, _ = trace_args
        module_name = frame.f_globals.get('__name__')
        return match_any(module_name, self._modules_to_trace)



import asyncio
import threading
from asyncio import Future, Task
from collections.abc import AsyncGenerator, AsyncIterator, Iterable
from threading import Thread
from typing import TypeVar


def current_task_or_thread() -> Task | Thread:
    """The asyncio task object if in a task, else the thread object"""
    try:
        task = asyncio.current_task()
    except RuntimeError:
        task = None
    return task or threading.current_thread()


T = TypeVar("T")


async def agen_with_wait(
    agen: AsyncIterator[T],
) -> AsyncGenerator[
    T | tuple[tuple[Future, ...], tuple[Future, ...]], Iterable[Future]
]:
    """Yield from the agen while waiting for received tasks

    Used to raise an exception from tasks
    """
    done = set[Future]()
    pending = set[Future]()
    anext = asyncio.ensure_future(agen.__anext__())
    while True:
        done_, pending_ = await asyncio.wait(
            pending | {anext},
            return_when=asyncio.FIRST_COMPLETED,
        )
        for t in done_ - {anext}:
            if exc := t.exception():
                anext.cancel()
                raise exc
        if anext in done_:
            try:
                item = anext.result()
            except StopAsyncIteration:
                break
            done_.remove(anext)
            done |= done_
        else:
            done |= done_
            continue
        pending &= pending_
        new = yield item
        if new is not None:
            pending |= set(new)
            yield tuple(done), tuple(pending)
            done.clear()
        anext = asyncio.ensure_future(agen.__anext__())


async def merge_aiters(
    *aiters: AsyncIterator[T],
) -> AsyncIterator[tuple[int, T]]:
    aiter_map = {a: i for i, a in enumerate(aiters)}
    task_map = {asyncio.ensure_future(a.__anext__()): a for a in aiter_map.keys()}
    tasks = set(task_map.keys())
    while tasks:
        done, pending = await asyncio.wait(
            tasks,
            return_when=asyncio.FIRST_COMPLETED,
        )
        tasks = pending
        while done:
            task = done.pop()
            try:
                item = task.result()
            except StopAsyncIteration:
                continue
            aiter = task_map[task]
            yield aiter_map[aiter], item
            task = asyncio.ensure_future(aiter.__anext__())
            tasks.add(task)
            task_map[task] = aiter



import dataclasses
from datetime import timezone
from typing import Optional

from nextline.plugin.spec import Context, hookimpl
from nextline.types import RunInfo


class RunInfoRegistrar:
    def __init__(self) -> None:
        self._run_info: Optional[RunInfo] = None

    @hookimpl
    async def on_initialize_run(self, context: Context) -> None:
        assert context.run_arg
        if isinstance(context.run_arg.statement, str):
            script = context.run_arg.statement
        else:
            script = None
        self._run_info = RunInfo(
            run_no=context.run_arg.run_no, state='initialized', script=script
        )
        await context.pubsub.publish('run_info', self._run_info)

    @hookimpl
    async def on_start_run(self, context: Context) -> None:
        assert self._run_info is not None
        assert context.running_process
        assert context.running_process.process_created_at.tzinfo is timezone.utc
        started_at = context.running_process.process_created_at.replace(tzinfo=None)
        self._run_info = dataclasses.replace(
            self._run_info, state='running', started_at=started_at
        )
        await context.pubsub.publish('run_info', self._run_info)

    @hookimpl
    async def on_end_run(self, context: Context) -> None:
        assert self._run_info is not None
        assert context.exited_process
        run_result = context.exited_process.returned
        assert run_result
        assert context.exited_process.process_exited_at.tzinfo is timezone.utc
        ended_at = context.exited_process.process_exited_at.replace(tzinfo=None)

        self._run_info = dataclasses.replace(
            self._run_info,
            state='finished',
            result=run_result.fmt_ret,
            exception=run_result.fmt_exc,
            ended_at=ended_at,
        )
        await context.pubsub.publish('run_info', self._run_info)
        self._run_info = None



__all__ = [
    '__version__',
    'disable_trace',
    'Nextline',
    'Statement',
]

from nextline.__about__ import __version__

from .disable import disable_trace
from .main import Nextline
from .types import Statement



import json
import traceback
from dataclasses import InitVar, dataclass, field
from queue import Queue
from types import FrameType
from typing import Any, Callable, Optional

from nextline.events import Event
from nextline.spawned.path import to_canonic_path
from nextline.types import RunNo, Statement

from .commands import Command

# if TYPE_CHECKING:
#     from sys import TraceFunction as TraceFunc  # type: ignore  # noqa: F401

# NOTE: TraceFunction from sys does not work well.
TraceFunction = Callable[[FrameType, str, Any], "TraceFunction | None"]

TraceArgs = tuple[FrameType, str, Any]

QueueIn = Queue[Command]
QueueOut = Queue[Event]


@dataclass
class RunArg:
    run_no: RunNo
    statement: Statement
    filename: Optional[str] = None
    trace_threads: bool = True
    trace_modules: bool = True


@dataclass
class RunResult:
    ret: Optional[Any] = None
    exc: InitVar[Optional[BaseException]] = None
    fmt_ret: Optional[str] = field(init=False, repr=False, default=None)
    fmt_exc: Optional[str] = field(init=False, repr=False, default=None)

    # NOTE: The `exc` is not stored because it is not always picklable, for
    # example, a dynamically defined exception class.

    def __post_init__(self, exc: Optional[BaseException]) -> None:
        self.fmt_ret = self._fmt_ret()
        self.fmt_exc = self._fmt_exc(exc)

    def _fmt_ret(self) -> str:
        return json.dumps(self.ret)

    def _fmt_exc(self, exc: Optional[BaseException]) -> str:
        if exc is None:
            return ''
        return ''.join(traceback.format_exception(type(exc), exc, exc.__traceback__))

    def result(self) -> Any:
        return self.ret


@dataclass
class TraceCallInfo:
    args: TraceArgs
    file_name: str = field(init=False)
    line_no: int = field(init=False)
    frame_object_id: int = field(init=False)
    event: str = field(init=False)

    def __post_init__(self) -> None:
        frame, event, _ = self.args
        self.file_name = to_canonic_path(frame.f_code.co_filename)
        self.line_no = frame.f_lineno
        self.frame_object_id = id(frame)
        self.event = event



import os
from typing import Callable


def ToCanonicPath() -> Callable[[str], str]:
    # Based on Bdb.canonic()
    # https://github.com/python/cpython/blob/v3.10.5/Lib/bdb.py#L39-L54

    cache = dict[str, str]()

    def _to_canonic_path(filename: str) -> str:
        if filename == "<" + filename[1:-1] + ">":
            return filename
        canonic = cache.get(filename)
        if not canonic:
            canonic = os.path.abspath(filename)
            canonic = os.path.normcase(canonic)
            cache[filename] = canonic
        return canonic

    return _to_canonic_path


to_canonic_path = ToCanonicPath()



import datetime
from dataclasses import dataclass
from typing import Optional

from nextline.types import PromptNo, RunNo, TaskNo, ThreadNo, TraceNo


@dataclass
class Event:
    pass


@dataclass
class OnStartTrace(Event):
    started_at: datetime.datetime
    run_no: RunNo
    trace_no: TraceNo
    thread_no: ThreadNo
    task_no: Optional[TaskNo]


@dataclass
class OnEndTrace(Event):
    ended_at: datetime.datetime
    run_no: RunNo
    trace_no: TraceNo


@dataclass
class OnStartTraceCall(Event):
    started_at: datetime.datetime
    run_no: RunNo
    trace_no: TraceNo
    file_name: str
    line_no: int
    frame_object_id: int
    event: str


@dataclass
class OnEndTraceCall(Event):
    ended_at: datetime.datetime
    run_no: RunNo
    trace_no: TraceNo


@dataclass
class OnStartCmdloop(Event):
    started_at: datetime.datetime
    run_no: RunNo
    trace_no: TraceNo


@dataclass
class OnEndCmdloop(Event):
    ended_at: datetime.datetime
    run_no: RunNo
    trace_no: TraceNo


@dataclass
class OnStartPrompt(Event):
    started_at: datetime.datetime
    run_no: RunNo
    trace_no: TraceNo
    prompt_no: PromptNo
    prompt_text: str
    file_name: str
    line_no: int
    frame_object_id: int
    event: str


@dataclass
class OnEndPrompt(Event):
    ended_at: datetime.datetime
    run_no: RunNo
    trace_no: TraceNo
    prompt_no: PromptNo
    command: str


@dataclass
class OnWriteStdout(Event):
    written_at: datetime.datetime
    run_no: RunNo
    trace_no: TraceNo
    text: str



import dataclasses
import datetime

from nextline.events import OnEndTrace, OnStartTrace
from nextline.plugin.spec import Context, hookimpl
from nextline.types import TraceInfo, TraceNo


class TraceInfoRegistrar:
    def __init__(self) -> None:
        self._trace_info_map = dict[TraceNo, TraceInfo]()

    @hookimpl
    async def on_initialize_run(self) -> None:
        self._trace_info_map = {}

    @hookimpl
    async def on_end_run(self, context: Context) -> None:
        while self._trace_info_map:
            # the process might have been killed.
            _, trace_info = self._trace_info_map.popitem()
            trace_info_end = dataclasses.replace(
                trace_info,
                state='finished',
                ended_at=datetime.datetime.utcnow(),
            )
            await context.pubsub.publish('trace_info', trace_info_end)

    @hookimpl
    async def on_start_trace(self, context: Context, event: OnStartTrace) -> None:
        assert context.run_arg
        trace_info = TraceInfo(
            run_no=context.run_arg.run_no,
            trace_no=event.trace_no,
            thread_no=event.thread_no,
            task_no=event.task_no,
            state='running',
            started_at=event.started_at,
        )
        self._trace_info_map[event.trace_no] = trace_info
        await context.pubsub.publish('trace_info', trace_info)

    @hookimpl
    async def on_end_trace(self, context: Context, event: OnEndTrace) -> None:
        trace_info = self._trace_info_map.pop(event.trace_no, None)
        if trace_info is None:
            # on_end_run() might have already been called
            return
        trace_info_end = dataclasses.replace(
            trace_info,
            state='finished',
            ended_at=event.ended_at,
        )
        await context.pubsub.publish('trace_info', trace_info_end)



import asyncio
import contextlib
import multiprocessing as mp
from collections.abc import AsyncIterator, Callable
from functools import partial
from logging import getLogger
from typing import Any, Optional, cast

import apluggy

from nextline import spawned
from nextline.plugin.spec import Context, hookimpl
from nextline.spawned import Command, QueueIn, QueueOut, RunResult
from nextline.utils import Timer, run_in_process


class RunSession:
    @hookimpl
    @apluggy.asynccontextmanager
    async def run(self, context: Context) -> AsyncIterator[None]:
        assert context.run_arg
        context.exited_process = None
        mp_context = mp.get_context('spawn')
        queue_in = cast(QueueIn, mp_context.Queue())
        queue_out = cast(QueueOut, mp_context.Queue())
        context.send_command = SendCommand(queue_in)
        async with relay_events(context, queue_out):
            context.running_process = await run_in_process(
                func=partial(spawned.main, context.run_arg),
                mp_context=mp_context,
                initializer=partial(spawned.set_queues, queue_in, queue_out),
                collect_logging=True,
            )
            await context.hook.ahook.on_start_run(context=context)
            try:
                yield
            finally:
                context.exited_process = await context.running_process
                if context.exited_process.returned is None:
                    context.exited_process.returned = RunResult()
                context.running_process = None
                if context.exited_process.raised:
                    logger = getLogger(__name__)
                    logger.exception(context.exited_process.raised)
        await context.hook.ahook.on_end_run(context=context)


def SendCommand(queue_in: QueueIn) -> Callable[[Command], None]:
    def _send_command(command: Command) -> None:
        logger = getLogger(__name__)
        logger.debug(f'send_pdb_command({command!r}')
        queue_in.put(command)

    return _send_command


@contextlib.asynccontextmanager
async def relay_events(context: Context, queue: QueueOut) -> AsyncIterator[None]:
    '''Call the hook `on_event_in_process()` on events emitted in the spawned process.'''
    logger = getLogger(__name__)

    in_finally = False
    timer = Timer(timeout=1)  # seconds

    async def _monitor() -> None:
        while (event := await asyncio.to_thread(queue.get)) is not None:
            logger.debug(f'event: {event!r}')
            await context.hook.ahook.on_event_in_process(context=context, event=event)
            if in_finally:
                timer.restart()

    task = asyncio.create_task(_monitor())
    try:
        yield
    finally:
        in_finally = True
        timer.restart()
        while not queue.empty():
            await asyncio.sleep(0)  # let _monitor() run
            if timer.is_timeout():
                logger.warning(f'Timeout. the queue is not empty: {queue!r}')
                break
        await asyncio.to_thread(queue.put, None)  # type: ignore
        await task


class Signal:
    @hookimpl
    async def interrupt(self, context: Context) -> None:
        assert context.running_process
        context.running_process.interrupt()

    @hookimpl
    async def terminate(self, context: Context) -> None:
        assert context.running_process
        context.running_process.terminate()

    @hookimpl
    async def kill(self, context: Context) -> None:
        assert context.running_process
        context.running_process.kill()


class CommandSender:
    @hookimpl
    async def send_command(self, context: Context, command: Command) -> None:
        assert context.send_command
        context.send_command(command)


class Result:
    @hookimpl
    def format_exception(self, context: Context) -> Optional[str]:
        if not context.exited_process:
            return None
        if not context.exited_process.returned:
            return None
        return context.exited_process.returned.fmt_exc

    @hookimpl
    def result(self, context: Context) -> Any:
        if not context.exited_process:
            return None
        if not context.exited_process.returned:
            return None
        return context.exited_process.returned.result()



from nextline.plugin.spec import Context, hookimpl


class StateNameRegistrar:
    @hookimpl
    async def on_change_state(self, context: Context, state_name: str) -> None:
        await context.pubsub.publish('state_name', state_name)



from collections.abc import Callable, Iterator, MutableMapping
from concurrent.futures import ThreadPoolExecutor
from logging import getLogger
from queue import Queue
from typing import TypeVar

from apluggy import PluginManager, contextmanager

from nextline.spawned.commands import PdbCommand
from nextline.spawned.plugin.spec import hookimpl
from nextline.spawned.types import QueueIn
from nextline.types import PromptNo, TraceNo

QueueMap = MutableMapping[TraceNo, 'Queue[PdbCommand]']


class Prompt:
    '''A plugin that responds to the hook prompt() with commands from a queue.'''

    def __init__(self) -> None:
        self._logger = getLogger(__name__)

    @hookimpl
    def init(self, hook: PluginManager, queue_in: QueueIn) -> None:
        self._hook = hook
        self._queue_in = queue_in
        self._queue_map: QueueMap = {}

    @hookimpl
    @contextmanager
    def context(self) -> Iterator[None]:
        with relay_commands(self._queue_in, self._queue_map):
            yield

    @hookimpl
    def on_start_trace(self, trace_no: TraceNo) -> None:
        self._queue_map[trace_no] = Queue()

    @hookimpl
    def on_end_trace(self, trace_no: TraceNo) -> None:
        del self._queue_map[trace_no]

    @hookimpl
    def prompt(self, prompt_no: PromptNo) -> str:
        trace_no = self._hook.hook.current_trace_no()
        self._logger.debug(f'PromptNo: {prompt_no}')
        queue = self._queue_map[trace_no]

        while True:
            pdb_command = queue.get()
            try:
                assert pdb_command.trace_no == trace_no
            except AssertionError:
                msg = f'TraceNo mismatch: {pdb_command.trace_no} != {trace_no}'
                self._logger.exception(msg)
                raise
            if not (n := pdb_command.prompt_no) == prompt_no:
                self._logger.warning(f'PromptNo mismatch: {n} != {prompt_no}')
                continue
            return pdb_command.command


@contextmanager
def relay_commands(queue_in: QueueIn, queue_map: QueueMap) -> Iterator[None]:
    '''Pass the Pdb commands from the main process to the Pdb instances.'''
    logger = getLogger(__name__)

    def fn() -> None:
        assert queue_in
        while msg := queue_in.get():
            logger.debug(f'queue_in.get() -> {msg!r}')
            if isinstance(msg, PdbCommand):
                queue_map[msg.trace_no].put(msg)

    with ThreadPoolExecutor(max_workers=1) as executor:
        future = executor.submit(try_again_on_error, fn)  # type: ignore
        try:
            yield
        finally:
            queue_in.put(None)  # type: ignore
            future.result()


_T = TypeVar('_T')


def try_again_on_error(func: Callable[[], _T]) -> _T:
    '''Keep trying until the function succeeds without an exception.'''
    while True:
        try:
            return func()
        # except KeyboardInterrupt:
        #     raise
        except BaseException:
            logger = getLogger(__name__)
            logger.exception('')



import datetime
from collections.abc import Generator, Iterator

from apluggy import PluginManager, contextmanager

from nextline.events import (
    OnEndCmdloop,
    OnEndPrompt,
    OnEndTrace,
    OnEndTraceCall,
    OnStartCmdloop,
    OnStartPrompt,
    OnStartTrace,
    OnStartTraceCall,
    OnWriteStdout,
)
from nextline.spawned.plugin.spec import hookimpl
from nextline.spawned.types import QueueOut, RunArg, TraceCallInfo
from nextline.types import PromptNo, TraceNo


class Repeater:
    @hookimpl
    def init(self, hook: PluginManager, run_arg: RunArg, queue_out: QueueOut) -> None:
        self._hook = hook
        self._run_no = run_arg.run_no
        self._queue_out = queue_out

    @hookimpl
    def on_start_trace(self, trace_no: TraceNo) -> None:
        assert trace_no == self._hook.hook.current_trace_no()
        started_at = datetime.datetime.utcnow()
        thread_no = self._hook.hook.current_thread_no()
        task_no = self._hook.hook.current_task_no()
        event = OnStartTrace(
            run_no=self._run_no,
            started_at=started_at,
            trace_no=trace_no,
            thread_no=thread_no,
            task_no=task_no,
        )
        self._queue_out.put(event)

    @hookimpl
    def on_end_trace(self, trace_no: TraceNo) -> None:
        ended_at = datetime.datetime.utcnow()
        event = OnEndTrace(ended_at=ended_at, run_no=self._run_no, trace_no=trace_no)
        self._queue_out.put(event)

    @hookimpl
    @contextmanager
    def on_trace_call(self, trace_call_info: TraceCallInfo) -> Iterator[None]:
        started_at = datetime.datetime.utcnow()
        trace_no = self._hook.hook.current_trace_no()
        event_start = OnStartTraceCall(
            started_at=started_at,
            run_no=self._run_no,
            trace_no=trace_no,
            file_name=trace_call_info.file_name,
            line_no=trace_call_info.line_no,
            frame_object_id=trace_call_info.frame_object_id,
            event=trace_call_info.event,
        )
        self._queue_out.put(event_start)

        try:
            yield
        finally:
            ended_at = datetime.datetime.utcnow()
            event_end = OnEndTraceCall(
                ended_at=ended_at, run_no=self._run_no, trace_no=trace_no
            )
            self._queue_out.put(event_end)

    @hookimpl
    @contextmanager
    def on_cmdloop(self) -> Generator[None, None, None]:
        started_at = datetime.datetime.utcnow()
        trace_no = self._hook.hook.current_trace_no()
        event_start = OnStartCmdloop(
            started_at=started_at, run_no=self._run_no, trace_no=trace_no
        )
        self._queue_out.put(event_start)

        try:
            yield
        finally:
            ended_at = datetime.datetime.utcnow()
            event_end = OnEndCmdloop(
                ended_at=ended_at, run_no=self._run_no, trace_no=trace_no
            )
            self._queue_out.put(event_end)

    @hookimpl
    @contextmanager
    def on_prompt(self, prompt_no: PromptNo, text: str) -> Generator[None, str, None]:
        started_at = datetime.datetime.utcnow()
        trace_no: TraceNo = self._hook.hook.current_trace_no()
        trace_call_info: TraceCallInfo = self._hook.hook.current_trace_call_info()
        event_start = OnStartPrompt(
            started_at=started_at,
            run_no=self._run_no,
            trace_no=trace_no,
            prompt_no=prompt_no,
            prompt_text=text,
            file_name=trace_call_info.file_name,
            line_no=trace_call_info.line_no,
            frame_object_id=trace_call_info.frame_object_id,
            event=trace_call_info.event,
        )
        self._queue_out.put(event_start)

        command = ''

        try:
            command = yield
            yield
        finally:
            ended_at = datetime.datetime.utcnow()
            event_end = OnEndPrompt(
                ended_at=ended_at,
                run_no=self._run_no,
                trace_no=trace_no,
                prompt_no=prompt_no,
                command=command,
            )
            self._queue_out.put(event_end)

    @hookimpl
    def on_write_stdout(self, trace_no: TraceNo, line: str) -> None:
        written_at = datetime.datetime.utcnow()
        trace_no = self._hook.hook.current_trace_no()
        event = OnWriteStdout(
            written_at=written_at,
            run_no=self._run_no,
            trace_no=trace_no,
            text=line,
        )
        self._queue_out.put(event)



import contextlib
import sys
from collections.abc import Iterator


@contextlib.contextmanager
def disable_trace() -> Iterator[None]:
    '''Remove the system trace function temporarily.

    Example:

    >>> with disable_trace():
    ...     # code without tracing
    ...     pass

    '''
    trace = sys.gettrace()
    sys.settrace(None)
    try:
        yield
    finally:
        sys.settrace(trace)



from typing import Optional

from nextline.events import OnWriteStdout
from nextline.plugin.spec import Context, hookimpl
from nextline.types import RunNo, StdoutInfo


class StdoutRegistrar:
    def __init__(self) -> None:
        self._run_no: Optional[RunNo] = None

    @hookimpl
    async def on_write_stdout(self, context: Context, event: OnWriteStdout) -> None:
        assert context.run_arg
        stdout_info = StdoutInfo(
            run_no=context.run_arg.run_no,
            trace_no=event.trace_no,
            text=event.text,
            written_at=event.written_at,
        )
        await context.pubsub.publish('stdout', stdout_info)



from nextline.plugin.spec import Context, hookimpl


class ScriptRegistrar:
    @hookimpl
    async def on_change_script(
        self, context: Context, script: str, filename: str
    ) -> None:
        await context.pubsub.publish('statement', script)
        await context.pubsub.publish('script_file_name', filename)



import threading
from asyncio import Task
from collections.abc import Iterator
from logging import getLogger
from threading import Thread
from typing import Optional
from weakref import WeakKeyDictionary, WeakSet

from apluggy import PluginManager, contextmanager

from nextline.count import TraceNoCounter
from nextline.spawned.plugin.spec import hookimpl
from nextline.types import TaskNo, ThreadNo, TraceNo
from nextline.utils import (
    ThreadTaskDoneCallback,
    ThreadTaskIdComposer,
    current_task_or_thread,
)


class TaskAndThreadKeeper:
    def __init__(self) -> None:
        self._set = WeakSet[Task | Thread]()
        self._counter = ThreadTaskIdComposer()
        self._main_thread: Optional[Thread] = None
        self._to_end: Optional[Thread] = None
        self._logger = getLogger(__name__)

    @hookimpl
    def init(self, hook: PluginManager) -> None:
        self._hook = hook

    @hookimpl
    @contextmanager
    def context(self) -> Iterator[None]:
        self._callback = ThreadTaskDoneCallback(done=self._on_end)
        self._main_thread = threading.current_thread()
        try:
            yield
        finally:
            self._callback.close()
            if self._to_end:
                self._on_end(self._to_end)

    @hookimpl
    def filtered(self) -> None:
        current = current_task_or_thread()
        if current not in self._set:
            self._on_start(current)
            self._set.add(current)

    def _on_start(self, current: Task | Thread) -> None:
        self._logger.info(f'{self.__class__.__name__}._on_start: {current}')
        if current is self._main_thread:
            self._to_end = self._main_thread
        else:
            self._callback.register(current)
        self._counter()  # increment the counter
        self._hook.hook.on_start_task_or_thread()

    def _on_end(self, ending: Task | Thread) -> None:
        # The "ending" is not the "current" unless it is the main thread.
        self._logger.info(f'{self.__class__.__name__}._on_end: {ending}')
        self._hook.hook.on_end_task_or_thread(task_or_thread=ending)

    @hookimpl
    def current_thread_no(self) -> ThreadNo:
        return self._counter().thread_no

    @hookimpl
    def current_task_no(self) -> Optional[TaskNo]:
        return self._counter().task_no


class TaskOrThreadToTraceMapper:
    def __init__(self) -> None:
        self._map = WeakKeyDictionary[Task | Thread, TraceNo]()
        self._counter = TraceNoCounter(1)
        self._logger = getLogger(__name__)

    @hookimpl
    def init(self, hook: PluginManager) -> None:
        self._hook = hook

    @hookimpl
    def on_start_task_or_thread(self) -> None:
        trace_no = self._counter()
        self._logger.info(f'{self.__class__.__name__} start: trace_no={trace_no}')
        self._map[current_task_or_thread()] = trace_no
        self._hook.hook.on_start_trace(trace_no=trace_no)

    @hookimpl
    def on_end_task_or_thread(self, task_or_thread: Task | Thread) -> None:
        trace_no = self._map[task_or_thread]
        self._hook.hook.on_end_trace(trace_no=trace_no)
        self._logger.info(f'{self.__class__.__name__} end: trace_no={trace_no}')

    @hookimpl
    def current_trace_no(self) -> Optional[TraceNo]:
        return self._map.get(current_task_or_thread())



import inspect
from types import FrameType
from typing import Optional

from apluggy import PluginManager

from .call import sys_trace
from .plugin import Hook
from .types import QueueIn, QueueOut, RunArg, RunResult


def run(run_arg: RunArg, queue_in: QueueIn, queue_out: QueueOut) -> RunResult:
    hook = Hook(run_arg=run_arg, queue_in=queue_in, queue_out=queue_out)
    with hook.with_.context():
        result = _compile_and_run(hook=hook, run_arg=run_arg)
        return result


def _compile_and_run(hook: PluginManager, run_arg: RunArg) -> RunResult:
    try:
        func = hook.hook.compose_callable()
    except BaseException as exc:
        _remove_frame(exc=exc, frame=inspect.currentframe())
        hook.hook.clean_exception(exc=exc)
        return RunResult(exc=exc)
    trace_func = hook.hook.create_trace_func()
    try:
        with sys_trace(trace_func=trace_func, thread=run_arg.trace_threads):
            ret = func()
        return RunResult(ret=ret)
    except BaseException as exc:
        _remove_frame(exc=exc, frame=inspect.currentframe())
        hook.hook.clean_exception(exc=exc)
        return RunResult(exc=exc)


def _remove_frame(exc: BaseException, frame: Optional[FrameType]) -> None:
    if exc.__traceback__ and frame and exc.__traceback__.tb_frame is frame:
        exc.__traceback__ = exc.__traceback__.tb_next



from contextlib import contextmanager

from nextline.spawned.call import sys_trace

MODULES_TO_SKIP = {
    'multiprocessing.*',
    'threading',
    'queue',
    'importlib',
    'asyncio.*',
    'codec',
    'concurrent.futures.*',
    'selectors',
    'weakref',
    '_weakrefset',
    'socket',
    'logging',
    'os',
    'collections.*',
    'importlib.*',
    'pathlib',
    'typing',
    'posixpath',
    'fnmatch',
    '_pytest.*',
    'apluggy.*',
    'pluggy.*',
    sys_trace.__module__,  # skip the 1st line of the finally clause in sys_trace()
    contextmanager.__module__,  # to skip contextlib.__exit__() in sys_trace()
}



from collections import defaultdict
from collections.abc import Callable, Iterator
from typing import Any, ContextManager, TypeVar

from apluggy import PluginManager, contextmanager

from nextline.spawned.plugin.spec import hookimpl
from nextline.types import TraceNo
from nextline.utils import peek_stdout


class PeekStdout:
    @hookimpl
    def init(self, hook: PluginManager) -> None:
        self._hook = hook

    @hookimpl
    def on_start_trace(self, trace_no: TraceNo) -> None:
        assert trace_no == self._key_factory()

    @hookimpl
    @contextmanager
    def context(self) -> Iterator[None]:
        with peek_stdout_by_key(key_factory=self._key_factory, callback=self._callback):
            yield

    def _key_factory(self) -> TraceNo | None:
        return self._hook.hook.current_trace_no()

    def _callback(self, trace_no: TraceNo, line: str) -> None:
        self._hook.hook.on_write_stdout(trace_no=trace_no, line=line)


_T = TypeVar('_T')


def peek_stdout_by_key(
    key_factory: Callable[[], _T | None],
    callback: Callable[[_T, str], Any],
) -> ContextManager[Callable[[str], int]]:
    callback_ = ReadLinesByKey(callback)
    assign_key = AssignKey(key_factory=key_factory, callback=callback_)
    return peek_stdout(assign_key)


def ReadLinesByKey(callback: Callable[[_T, str], Any]) -> Callable[[_T, str], Any]:
    buffer = defaultdict[_T, str](str)

    def read_lines_by_key(key: _T, s: str) -> None:
        buffer[key] += s
        if s.endswith('\n'):
            line = buffer.pop(key)
            callback(key, line)

    return read_lines_by_key


def AssignKey(
    key_factory: Callable[[], _T | None], callback: Callable[[_T, str], Any]
) -> Callable[[str], None]:
    def assign_key(s: str) -> None:
        if key := key_factory():
            callback(key, s)

    return assign_key



from logging import getLogger
from typing import Optional

from nextline.events import OnEndTraceCall, OnStartPrompt, OnStartTraceCall
from nextline.plugin.spec import Context, hookimpl
from nextline.types import PromptNotice, RunNo, TraceNo


class PromptNoticeRegistrar:
    def __init__(self) -> None:
        self._run_no: Optional[RunNo] = None
        self._trace_call_map = dict[TraceNo, OnStartTraceCall]()
        self._logger = getLogger(__name__)

    @hookimpl
    async def on_initialize_run(self) -> None:
        self._trace_call_map.clear()

    @hookimpl
    async def on_end_run(self, context: Context) -> None:
        await context.pubsub.end('prompt_notice')

    @hookimpl
    async def on_start_trace_call(self, event: OnStartTraceCall) -> None:
        self._trace_call_map[event.trace_no] = event

    @hookimpl
    async def on_end_trace_call(self, event: OnEndTraceCall) -> None:
        self._trace_call_map.pop(event.trace_no, None)

    @hookimpl
    async def on_start_prompt(self, context: Context, event: OnStartPrompt) -> None:
        assert context.run_arg
        trace_no = event.trace_no
        prompt_no = event.prompt_no
        trace_call = self._trace_call_map[trace_no]
        prompt_notice = PromptNotice(
            started_at=event.started_at,
            run_no=context.run_arg.run_no,
            trace_no=trace_no,
            prompt_no=prompt_no,
            prompt_text=event.prompt_text,
            event=trace_call.event,
            file_name=trace_call.file_name,
            line_no=trace_call.line_no,
        )
        await context.pubsub.publish('prompt_notice', prompt_notice)



import sys
from pathlib import Path
from types import CodeType
from typing import Any, Callable

from nextline.spawned.path import to_canonic_path
from nextline.spawned.plugin.spec import hookimpl
from nextline.spawned.types import RunArg

from . import _script


class CallableComposer:
    @hookimpl
    def init(self, run_arg: RunArg) -> None:
        self._run_arg = run_arg

    @hookimpl
    def compose_callable(self) -> Callable[[], Any]:
        return _compose_callable(self._run_arg)

    @hookimpl
    def clean_exception(self, exc: BaseException) -> None:
        if not isinstance(exc, SyntaxError):
            return
        is_compiled_here = False
        tb = exc.__traceback__
        while tb:
            module = tb.tb_frame.f_globals.get('__name__')
            if module == __name__:
                is_compiled_here = True
            if not tb.tb_next and is_compiled_here:
                # NOTE: tb.tb_next is None. No traceback for SyntaxError.
                exc.__traceback__ = tb.tb_next
            tb = tb.tb_next


def _compose_callable(run_arg: RunArg) -> Callable[[], Any]:
    # TODO: Rewrite with a match statement for Python 3.10
    statement = run_arg.statement
    filename = run_arg.filename

    if isinstance(statement, str):
        assert filename is not None
        code = compile(statement, filename, 'exec')
        return _script.compose(code)
    elif isinstance(statement, Path):
        return _from_path(statement)
    elif isinstance(statement, CodeType):
        return _script.compose(statement)
    elif callable(statement):
        return statement
    else:
        raise TypeError(f'statement: {statement!r}')


def _from_path(path: Path) -> Callable[[], Any]:
    # Read as a str and compile it as Pdb does.
    # https://github.com/python/cpython/blob/v3.10.10/Lib/pdb.py#L1568-L1592
    path = Path(to_canonic_path(str(path)))
    statement = path.read_text()
    code = compile(statement, str(path), 'exec')
    sys.path.insert(0, str(path.parent))
    return _script.compose(code)



import time


class Timer:
    '''A timer.

    Parameters
    ----------
    timeout :
        The timeout in seconds.


    Examples
    --------
    >>> timer = Timer(timeout=0.01)
    >>> timer.elapsed() < 0.01
    True

    >>> timer.is_timeout()
    False

    >>> timer.restart()

    >>> timer.is_timeout()
    False

    >>> time.sleep(0.01)

    >>> timer.is_timeout()
    True


    '''

    def __init__(self, timeout: float) -> None:
        self._timeout = timeout
        self._start = time.perf_counter()

    def restart(self) -> None:
        '''Restart the timer.'''
        self._start = time.perf_counter()

    def elapsed(self) -> float:
        '''Return the time passed in seconds.'''
        return time.perf_counter() - self._start

    def is_timeout(self) -> bool:
        '''Return True if the timeout is reached.'''
        return self.elapsed() > self._timeout



import asyncio
from logging import getLogger
from typing import Any, Optional

from transitions import EventData

from nextline.plugin import Context
from nextline.spawned import Command
from nextline.types import ResetOptions

from .factory import build_state_machine


class Machine:
    '''The finite state machine of the nextline states.'''

    def __init__(self, context: Context) -> None:
        self._context = context
        self._hook = context.hook
        self._machine = build_state_machine(model=self)
        self._machine.after_state_change = self.after_state_change.__name__  # type: ignore
        assert self.state  # type: ignore

    def __repr__(self) -> str:
        # e.g., "<Machine 'running'>"
        return f'<{self.__class__.__name__} {self.state!r}>'  # type: ignore

    async def after_state_change(self, event: EventData) -> None:
        if not (event.transition and event.transition.dest):
            # internal transition
            return
        await self._hook.ahook.on_change_state(
            context=self._context, state_name=self.state  # type: ignore
        )

    async def on_exit_created(self, _: EventData) -> None:
        await self._hook.ahook.start(context=self._context)

    async def on_enter_initialized(self, _: EventData) -> None:
        self._context.run_arg = self._hook.hook.compose_run_arg(context=self._context)
        await self._hook.ahook.on_initialize_run(context=self._context)

    async def on_enter_running(self, _: EventData) -> None:
        self.run_finished = asyncio.Event()
        run_started = asyncio.Event()

        async def run() -> None:
            async with self._hook.awith.run(context=self._context):
                run_started.set()
            self._context.run_arg = None
            await self.finish()  # type: ignore
            self.run_finished.set()

        self._task = asyncio.create_task(run())
        await run_started.wait()

    async def send_command(self, command: Command) -> None:
        await self._hook.ahook.send_command(context=self._context, command=command)

    async def interrupt(self) -> None:
        await self._hook.ahook.interrupt(context=self._context)

    async def terminate(self) -> None:
        await self._hook.ahook.terminate(context=self._context)

    async def kill(self) -> None:
        await self._hook.ahook.kill(context=self._context)

    async def on_close_while_running(self, _: EventData) -> None:
        await self.run_finished.wait()

    async def wait(self) -> None:
        await self.run_finished.wait()
    
    async def on_enter_finished(self, _: EventData) -> None:
        await self._hook.ahook.on_finished(context=self._context)

    async def on_exit_finished(self, _: EventData) -> None:
        await self._task

    def format_exception(self) -> Optional[str]:
        return self._hook.hook.format_exception(context=self._context)

    def result(self) -> Any:
        return self._hook.hook.result(context=self._context)

    async def on_enter_closed(self, _: EventData) -> None:
        await self._hook.ahook.close(context=self._context)

    async def on_reset(self, event: EventData) -> None:
        logger = getLogger(__name__)
        if args := list(event.args):
            logger.warning(f'Unexpected args: {args!r}')
        kwargs = event.kwargs
        reset_options: ResetOptions = kwargs.pop('reset_options')
        if kwargs:
            logger.warning(f'Unexpected kwargs: {kwargs!r}')
        await self._hook.ahook.reset(context=self._context, reset_options=reset_options)

    async def __aenter__(self) -> 'Machine':
        await self.initialize()  # type: ignore
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:  # type: ignore
        del exc_type, exc_value, traceback
        await self.close()  # type: ignore



from logging import getLogger
from typing import Any, Callable, ContextManager

from apluggy import PluginManager

from nextline.count import PromptNoCounter
from nextline.spawned.exc import NotOnTraceCall
from nextline.spawned.plugin.spec import hookimpl
from nextline.spawned.types import TraceFunction

from .custom import CustomizedPdb
from .stream import PromptFuncType, StdInOut


class PdbInstanceFactory:
    '''A plugin that creates local trace functions.

    Each time the first result only hook `create_local_trace_func` is called,
    this plugin creates a new instance of the customized Pdb class and returns
    its trace function.

    The hook `create_local_trace_func` is called for each async task or thread.
    '''

    @hookimpl
    def init(self, hook: PluginManager) -> None:
        self._factory = Factory(hook=hook)

    @hookimpl
    def create_local_trace_func(self) -> TraceFunction:
        return self._factory()


def Factory(hook: PluginManager) -> Callable[[], TraceFunction]:
    cmdloop_hook = CmdloopHook(hook=hook)
    prompt_func = PromptFunc(hook=hook)

    def _factory() -> TraceFunction:
        stdio = StdInOut(prompt_func=prompt_func)
        pdb = CustomizedPdb(
            cmdloop_hook=cmdloop_hook,
            stdin=stdio,
            stdout=stdio,
        )
        stdio.prompt_end = pdb.prompt
        return pdb.trace_dispatch

    return _factory


def CmdloopHook(hook: PluginManager) -> Callable[[], ContextManager[Any]]:
    '''Return a context manager in which Pdb.cmdloop() is called.'''

    def cmdloop() -> ContextManager[Any]:
        if not hook.hook.is_on_trace_call():
            raise NotOnTraceCall
        return hook.with_.on_cmdloop()

    return cmdloop


def PromptFunc(hook: PluginManager) -> PromptFuncType:
    '''Return a function that responds to Pdb prompts with user commands.'''

    counter = PromptNoCounter(1)
    logger = getLogger(__name__)

    def _prompt_func(text: str) -> str:
        prompt_no = counter()
        logger.debug(f'PromptNo: {prompt_no}')
        with (context := hook.with_.on_prompt(prompt_no=prompt_no, text=text)):
            command = hook.hook.prompt(prompt_no=prompt_no, text=text)
            if command is None:
                logger.warning(f'command is None: {command!r}')
            context.gen.send(command)
        return command

    return _prompt_func



from types import FrameType
from typing import Any, Callable, ContextManager, Optional

from .types import TraceFunction


def WithContext(
    trace: TraceFunction, context: Callable[[FrameType, str, Any], ContextManager[None]]
) -> TraceFunction:
    '''Return a new trace func that calls the given trace func in the context.'''

    def _create_local_trace() -> TraceFunction:
        next_trace: TraceFunction | None = trace

        def _local_trace(
            frame: FrameType, event: str, arg: Any
        ) -> Optional[TraceFunction]:
            nonlocal next_trace
            assert next_trace
            with context(frame, event, arg):
                if next_trace := next_trace(frame, event, arg):
                    return _local_trace
                return None

        return _local_trace

    def _global_trace(
        frame: FrameType, event: str, arg: Any
    ) -> Optional[TraceFunction]:
        return _create_local_trace()(frame, event, arg)

    return _global_trace



import sys
from collections.abc import Iterator
from contextlib import contextmanager
from typing import Any, Callable, ContextManager, TextIO


def peek_stdout(callback: Callable[[str], Any]) -> ContextManager[Callable[[str], int]]:
    '''A context manager that executes the callback with text written to stdout.

    Example:

    Collect arguments to the callback in this list:
    >>> written = []

    Define a callback that appends the argument to the list:
    >>> def callback(s: str) -> None:
    ...     written.append(s)

    Write to stdout in a context:
    >>> with peek_stdout(callback):
    ...     print('hello')
    ...     print('world')
    hello
    world

    The callback was called with the text written to stdout:
    >>> print(''.join(written), end='')
    hello
    world


    NOTE: Do not write to stdout in the callback. This will cause an infinite recursion.

    '''
    return peek_textio(sys.stdout, callback)


def peek_stderr(callback: Callable[[str], Any]) -> ContextManager[Callable[[str], int]]:
    '''A context manager that executes the callback with text written to stderr.'''
    return peek_textio(sys.stderr, callback)


@contextmanager
def peek_textio(
    textio: TextIO, callback: Callable[[str], Any]
) -> Iterator[Callable[[str], int]]:
    '''A context manager that executes the callback with text written to the textio.'''

    org_write = textio.write

    def write(s: str, /) -> int:
        callback(s)
        return org_write(s)

    textio.write = write  # type: ignore

    try:
        yield write
    finally:
        textio.write = org_write  # type: ignore



import asyncio
import linecache
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from logging import getLogger
from typing import Any, Optional

from .continuous import Continuous
from .fsm import Machine
from .plugin import Context, build_hook, log_loaded_plugins
from .spawned import PdbCommand
from .types import (
    InitOptions,
    PromptInfo,
    PromptNo,
    PromptNotice,
    ResetOptions,
    RunInfo,
    Statement,
    StdoutInfo,
    TraceInfo,
    TraceNo,
)
from .utils import merge_aiters
from .utils.pubsub.broker import PubSub


class Nextline:
    '''Nextline allows line-by-line execution of concurrent Python scripts

    Nextline supports concurrency with threading and asyncio. It uses multiple
    instances of Pdb, one for each thread and async task.


    Parameters
    ----------
    statement : str or Path or CodeType or Callable[[], Any]
        A Python code as a str, a Path object that points to a Python script,
        a CodeType object, or a callable with no arguments. It must be
        picklable.
    run_no_start_from : int, optional
        The first run number. The default is 1.
    trace_threads : bool, optional
        The default is False. If False, trace only the main thread. If True,
        trace all threads.
    trace_modules : bool, optional
        The default is False. If False, trace only the statement. If True,
        trace imported modules as well.
    timeout_on_exit : float, optional
        The timeout in seconds to wait for the nextline to exit from the "with"
        block. The default is 3.

    '''

    def __init__(
        self,
        statement: Statement,
        run_no_start_from: int = 1,
        trace_threads: bool = False,
        trace_modules: bool = False,
        timeout_on_exit: float = 3,
    ):
        self._init_options = InitOptions(
            statement=statement,
            run_no_start_from=run_no_start_from,
            trace_threads=trace_threads,
            trace_modules=trace_modules,
        )
        self._continuous = Continuous(self)
        self._pubsub = PubSub[Any, Any]()
        self._timeout_on_exit = timeout_on_exit
        self._started = False
        self._closed = False
        self._hook = build_hook()

    def __repr__(self) -> str:
        # e.g., "<Nextline 'running'>"
        return f'<{self.__class__.__name__} {self.state!r}>'

    def register(self, plugin: Any) -> str | None:
        return self._hook.register(plugin)

    async def start(self) -> None:
        if self._started:
            return
        self._started = True
        logger = getLogger(__name__)
        logger.debug(f'self._init_options: {self._init_options}')
        log_loaded_plugins(hook=self._hook)
        context = Context(nextline=self, hook=self._hook, pubsub=self._pubsub)
        self._hook.hook.init(context=context, init_options=self._init_options)
        self._machine = Machine(context=context)
        await self._continuous.start()
        await self._machine.initialize()  # type: ignore

    async def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        await self._pubsub.close()
        await self._machine.close()  # type: ignore
        await self._continuous.close()

    async def __aenter__(self) -> 'Nextline':
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:  # type: ignore
        del exc_type, exc_value, traceback
        await asyncio.wait_for(self.close(), timeout=self._timeout_on_exit)

    async def run(self) -> None:
        '''Start the script execution.'''
        await self._machine.run()  # type: ignore

    @asynccontextmanager
    async def run_session(self) -> AsyncIterator['Nextline']:
        '''Yield when the script execution is started and exit when it has finished.'''
        await self._machine.run()  # type: ignore
        try:
            yield self
        finally:
            await self._machine.wait()

    async def run_and_continue(self) -> None:
        '''Start the script execution in the non-interactive mode.

        Returns when the run has started.
        '''
        await self._continuous.run_and_continue()

    async def run_continue_and_wait(
        self, started: Optional[asyncio.Event] = None
    ) -> None:
        '''Start the script execution in the non-interactive mode and wait until end.

        Similar to run_and_continue() but returns when the run has finished.
        The event started is set when the run has started.
        '''
        started = started or asyncio.Event()
        await self._continuous.run_continue_and_wait(started)

    async def send_pdb_command(
        self, command: str, prompt_no: int, trace_no: int
    ) -> None:
        logger = getLogger(__name__)
        logger.debug(f'send_pdb_command({command!r}, {prompt_no!r}, {trace_no!r})')
        item = PdbCommand(
            trace_no=TraceNo(trace_no),
            prompt_no=PromptNo(prompt_no),
            command=command,
        )
        await self._machine.send_command(item)

    async def interrupt(self) -> None:
        await self._machine.interrupt()

    async def terminate(self) -> None:
        await self._machine.terminate()

    async def kill(self) -> None:
        await self._machine.kill()  # type: ignore

    def format_exception(self) -> Optional[str]:
        '''Formatted uncaught exception from the last run'''
        return self._machine.format_exception()

    def result(self) -> Any:
        '''Return value of the last run. None unless the statement is a callable.'''
        return self._machine.result()

    async def reset(
        self,
        statement: Optional[Statement] = None,
        run_no_start_from: Optional[int] = None,
        trace_threads: Optional[bool] = None,
        trace_modules: Optional[bool] = None,
    ) -> None:
        """Prepare for the next run"""
        reset_options = ResetOptions(
            statement=statement,
            run_no_start_from=run_no_start_from,
            trace_threads=trace_threads,
            trace_modules=trace_modules,
        )
        logger = getLogger(__name__)
        logger.debug(f'reset_options: {reset_options}')
        await self._machine.reset(  # type: ignore
            reset_options=reset_options,
        )

    @property
    def statement(self) -> str:
        """The script"""
        return self.get("statement")

    @property
    def state(self) -> Optional[str]:
        """The current condition of the script execution.

        The possible values are "initialized", "running", "finished",
        "closed"
        """
        try:
            return self._machine.state  # type: ignore
        except AttributeError:
            return None

    def subscribe_state(self) -> AsyncIterator[str]:
        return self.subscribe("state_name")

    @property
    def run_no(self) -> int:
        """The current run number"""
        return self.get("run_no")

    def subscribe_run_no(self) -> AsyncIterator[int]:
        return self.subscribe("run_no")

    @property
    def trace_ids(self) -> tuple[int, ...]:
        try:
            return self.get("trace_nos")
        except (KeyError, LookupError):
            return ()

    def subscribe_trace_ids(self) -> AsyncIterator[tuple[int, ...]]:
        return self.subscribe("trace_nos")

    def get_source(self, file_name: Optional[str] = None) -> list[str]:
        if not file_name or file_name == self._pubsub.latest("script_file_name"):
            return self.get("statement").split("\n")
        return [e.rstrip() for e in linecache.getlines(file_name)]

    def get_source_line(self, line_no: int, file_name: Optional[str] = None) -> str:
        """
        based on linecache.getline()
        https://github.com/python/cpython/blob/v3.9.5/Lib/linecache.py#L26
        """
        lines = self.get_source(file_name)
        if 1 <= line_no <= len(lines):
            return lines[line_no - 1]
        return ""

    def subscribe_run_info(self) -> AsyncIterator[RunInfo]:
        return self.subscribe("run_info")

    def subscribe_trace_info(self) -> AsyncIterator[TraceInfo]:
        return self.subscribe("trace_info")

    def prompts(self) -> AsyncIterator[PromptNotice]:
        '''Yield for each prompt. Return when the run ends.'''
        return self.subscribe('prompt_notice')

    def subscribe_prompt_info(self) -> AsyncIterator[PromptInfo]:
        return self.subscribe("prompt_info")

    def subscribe_prompt_info_for(self, trace_no: int) -> AsyncIterator[PromptInfo]:
        return self.subscribe(f"prompt_info_{trace_no}")

        # Alternative implementation under development
        # return self._subscribe_prompt_info_for(trace_no)

    async def _subscribe_prompt_info_for(
        self, trace_no: int
    ) -> AsyncIterator[PromptInfo]:
        merged = merge_aiters(
            self.subscribe_prompt_info(),
            self.subscribe_trace_info(),
        )
        async for _, info in merged:
            if not info.trace_no == trace_no:  # type: ignore
                continue
            if isinstance(info, TraceInfo):
                if info.trace_no == trace_no:
                    if info.state == "finished":
                        break
                continue
            assert isinstance(info, PromptInfo)
            yield info

    def get(self, key: Any) -> Any:
        return self._pubsub.latest(key)

    def subscribe(self, key: Any, last: Optional[bool] = True) -> AsyncIterator[Any]:
        return self._pubsub.subscribe(key, last=last)

    def subscribe_stdout(self) -> AsyncIterator[StdoutInfo]:
        return self.subscribe("stdout", last=False)

    @property
    def continuous_enabled(self) -> bool:
        return self._continuous.enabled

    def subscribe_continuous_enabled(self) -> AsyncIterator[bool]:
        return self._continuous.subscribe_enabled()



import dataclasses
from typing import TYPE_CHECKING, Any, Callable, Optional

import apluggy

from nextline import events, spawned
from nextline.types import InitOptions, ResetOptions
from nextline.utils import ExitedProcess, RunningProcess
from nextline.utils.pubsub.broker import PubSub

if TYPE_CHECKING:
    from nextline import Nextline

PROJECT_NAME = 'nextline_main'


hookspec = apluggy.HookspecMarker(PROJECT_NAME)
hookimpl = apluggy.HookimplMarker(PROJECT_NAME)


@dataclasses.dataclass
class Context:
    nextline: 'Nextline'
    hook: apluggy.PluginManager
    pubsub: PubSub
    run_arg: spawned.RunArg | None = None
    send_command: Callable[[spawned.Command], None] | None = None
    running_process: RunningProcess[spawned.RunResult] | None = None
    exited_process: ExitedProcess[spawned.RunResult] | None = None


@hookspec
def init(context: Context, init_options: InitOptions) -> None:
    pass


@hookspec
async def start(context: Context) -> None:
    pass


@hookspec
async def close(
    context: Context, exc_type=None, exc_value=None, traceback=None
) -> None:
    pass


@hookspec
async def reset(context: Context, reset_options: ResetOptions) -> None:
    pass


@hookspec
async def on_change_state(context: Context, state_name: str) -> None:
    pass


@hookspec
async def on_change_script(context: Context, script: str, filename: str) -> None:
    pass


@hookspec(firstresult=True)
def compose_run_arg(context: Context) -> Optional[spawned.RunArg]:
    pass


@hookspec
async def on_initialize_run(context: Context) -> None:
    pass


@hookspec
@apluggy.asynccontextmanager
async def run(context: Context):  # type: ignore
    yield


@hookspec
async def on_event_in_process(context: Context, event: events.Event) -> None:
    pass


@hookspec
async def on_start_run(context: Context) -> None:
    pass


@hookspec
async def interrupt(context: Context) -> None:
    pass


@hookspec
async def terminate(context: Context) -> None:
    pass


@hookspec
async def kill(context: Context) -> None:
    pass


@hookspec
async def send_command(context: Context, command: spawned.Command) -> None:
    pass


@hookspec
async def on_end_run(context: Context) -> None:
    '''The run is about to finish. The state is still 'running'.'''
    pass


@hookspec
async def on_finished(context: Context) -> None:
    '''The run has finished. The state is 'finished'.'''
    pass


@hookspec(firstresult=True)
def format_exception(context: Context) -> Optional[str]:
    pass


@hookspec(firstresult=True)
def result(context: Context) -> Any:
    pass


@hookspec
async def on_start_trace(context: Context, event: events.OnStartTrace) -> None:
    pass


@hookspec
async def on_end_trace(context: Context, event: events.OnEndTrace) -> None:
    pass


@hookspec
async def on_start_trace_call(
    context: Context, event: events.OnStartTraceCall
) -> None:
    pass


@hookspec
async def on_end_trace_call(context: Context, event: events.OnEndTraceCall) -> None:
    pass


@hookspec
async def on_start_cmdloop(context: Context, event: events.OnStartCmdloop) -> None:
    pass


@hookspec
async def on_end_cmdloop(context: Context, event: events.OnEndCmdloop) -> None:
    pass


@hookspec
async def on_start_prompt(context: Context, event: events.OnStartPrompt) -> None:
    pass


@hookspec
async def on_end_prompt(context: Context, event: events.OnEndPrompt) -> None:
    pass


@hookspec
async def on_write_stdout(context: Context, event: events.OnWriteStdout) -> None:
    pass



import time
from queue import Queue
from typing import Optional

from .timer import Timer


def wait_until_queue_empty(
    queue: Queue, timeout: Optional[float] = None, interval: float = 0.001
) -> None:
    '''Wait until the queue is empty.

    Parameters
    ----------
    queue :
        The queue to wait for.
    timeout : optional
        The timeout in seconds. Wait indefinitely if None.
    interval : float, optional
        The interval in seconds to check the queue.

    '''
    if timeout is None:
        while not queue.empty():
            time.sleep(interval)
    else:
        timer = Timer(timeout)
        while not queue.empty():
            if timer.is_timeout():
                raise TimeoutError(f'Timeout. the queue is not empty: {queue!r}')
            time.sleep(interval)



from nextline.count import RunNoCounter
from nextline.plugin.spec import Context, hookimpl
from nextline.spawned import RunArg
from nextline.types import InitOptions, ResetOptions

SCRIPT_FILE_NAME = "<string>"


class RunArgComposer:
    @hookimpl
    def init(self, init_options: InitOptions) -> None:
        self._run_no_count = RunNoCounter(init_options.run_no_start_from)
        self._statement = init_options.statement
        self._filename = SCRIPT_FILE_NAME
        self._trace_threads = init_options.trace_threads
        self._trace_modules = init_options.trace_modules

    @hookimpl
    async def start(self, context: Context) -> None:
        await context.hook.ahook.on_change_script(
            context=context, script=self._statement, filename=self._filename
        )

    @hookimpl
    async def reset(self, context: Context, reset_options: ResetOptions) -> None:
        if (statement := reset_options.statement) is not None:
            self._statement = statement
            await context.hook.ahook.on_change_script(
                context=context, script=self._statement, filename=self._filename
            )
        if (run_no_start_from := reset_options.run_no_start_from) is not None:
            self._run_no_count = RunNoCounter(run_no_start_from)
        if (trace_threads := reset_options.trace_threads) is not None:
            self._trace_threads = trace_threads
        if (trace_modules := reset_options.trace_modules) is not None:
            self._trace_modules = trace_modules

    @hookimpl
    def compose_run_arg(self) -> RunArg:
        run_arg = RunArg(
            run_no=self._run_no_count(),
            statement=self._statement,
            filename=self._filename,
            trace_threads=self._trace_threads,
            trace_modules=self._trace_modules,
        )
        return run_arg
